from __future__ import annotations

import asyncio
import os
from collections import abc
from typing import AsyncGenerator

import pytest
from piccolo.conf.apps import Finder
from piccolo.table import create_db_tables, drop_db_tables

os.environ["PICCOLO_CONF"] = "tests.piccolo_conf"

pytestmark = pytest.mark.anyio


@pytest.fixture(scope="session")
def anyio_backend() -> str:
    return "asyncio"


@pytest.fixture(scope="session")
def event_loop() -> abc.Iterator[asyncio.AbstractEventLoop]:
    """Scoped Event loop.

    Need the event loop scoped to the session so that we can use it to check
    containers are ready in session scoped containers fixture.
    """
    policy = asyncio.get_event_loop_policy()
    loop = policy.new_event_loop()
    try:
        yield loop
    finally:
        loop.close()


@pytest.fixture(autouse=True, scope="session")
async def scaffold_piccolo() -> AsyncGenerator:
    """Scaffolds Piccolo ORM and performs cleanup."""
    tables = Finder().get_table_classes()
    await drop_db_tables(*tables)
    await create_db_tables(*tables)
    yield
    await drop_db_tables(*tables)



from __future__ import annotations  # noqa: INP001

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from docutils.nodes import Element
    from sphinx.addnodes import pending_xref
    from sphinx.application import Sphinx
    from sphinx.environment import BuildEnvironment


def on_missing_reference(app: Sphinx, env: BuildEnvironment, node: pending_xref, contnode: Element) -> Any:
    if not hasattr(node, "attributes"):
        return None

    attributes = node.attributes
    target = attributes["reftarget"]
    py_domain = env.domains["py"]

    # autodoc sometimes incorrectly resolves these types, so we try to resolve them as py:data fist and fall back to any
    new_node = py_domain.resolve_xref(env, node["refdoc"], app.builder, "data", target, node, contnode)
    if new_node is None:
        resolved_xrefs = py_domain.resolve_any_xref(env, node["refdoc"], app.builder, target, node, contnode)
        for ref in resolved_xrefs:
            if ref:
                return ref[1]
    return new_node


def setup(app: Sphinx) -> dict[str, bool]:
    app.connect("missing-reference", on_missing_reference)
    app.add_config_value("ignore_missing_refs", default={}, rebuild=False)

    return {"parallel_read_safe": True, "parallel_write_safe": True}



from __future__ import annotations

from litestar_piccolo.dto import PiccoloDTO
from litestar_piccolo.plugin import PiccoloPlugin, PiccoloSerializationPlugin

__all__ = ("PiccoloPlugin", "PiccoloSerializationPlugin", "PiccoloDTO")



"""The contents of this file were adapted from:

https://github.com/piccolo-orm/piccolo/blob/master/tests/example_apps/music/piccolo_app.py
"""

from pathlib import Path

from piccolo.conf.apps import AppConfig

from tests.app.tables import (
    Band,
    Concert,
    Manager,
    RecordingStudio,
    Venue,
)

CURRENT_DIRECTORY = Path(__file__).parent

APP_CONFIG = AppConfig(
    app_name="music",
    table_classes=[
        Manager,
        Band,
        Venue,
        Concert,
        RecordingStudio,
    ],
    migrations_folder_path=str(CURRENT_DIRECTORY / "piccolo_migrations"),
    commands=[],
)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar

from litestar.plugins import InitPluginProtocol, SerializationPluginProtocol
from piccolo.table import Table

from litestar_piccolo.dto import PiccoloDTO

if TYPE_CHECKING:
    from litestar.config.app import AppConfig
    from litestar.typing import FieldDefinition

T = TypeVar("T", bound=Table)


class PiccoloSerializationPlugin(SerializationPluginProtocol):
    def __init__(self) -> None:
        self._type_dto_map: dict[type[Table], type[PiccoloDTO[Any]]] = {}

    def supports_type(self, field_definition: FieldDefinition) -> bool:
        return (
            field_definition.is_collection and field_definition.has_inner_subclass_of(Table)
        ) or field_definition.is_subclass_of(Table)

    def create_dto_for_type(self, field_definition: FieldDefinition) -> type[PiccoloDTO[Any]]:
        # assumes that the type is a container of Piccolo models or a single Piccolo model
        annotation = next(
            (inner_type.annotation for inner_type in field_definition.inner_types if inner_type.is_subclass_of(Table)),
            field_definition.annotation,
        )
        if annotation in self._type_dto_map:
            return self._type_dto_map[annotation]

        self._type_dto_map[annotation] = dto_type = PiccoloDTO[annotation]  # type:ignore[valid-type]

        return dto_type


class PiccoloPlugin(InitPluginProtocol):
    """A plugin that provides Piccolo integration."""

    def __init__(self) -> None:
        """Initialize ``PiccoloPlugin``."""

    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        """Configure application for use with Piccolo.

        Args:
            app_config: The :class:`AppConfig <.config.app.AppConfig>` instance.
        """
        app_config.plugins.extend([PiccoloSerializationPlugin()])
        return app_config



"""The contents of this file were adapted from:

https://github.com/piccolo-orm/piccolo/blob/master/tests/example_apps/music/tables.py
"""

from piccolo.columns.column_types import JSON, JSONB, Array, ForeignKey, Integer, Varchar
from piccolo.table import Table


class RecordingStudio(Table):
    facilities = JSON()
    facilities_b = JSONB()
    microphones = Array(Varchar())


class Manager(Table):
    name = Varchar(length=50)


class Band(Table):
    name = Varchar(length=50)
    manager = ForeignKey(Manager)
    popularity = Integer()


class Venue(Table):
    name = Varchar(length=100)
    capacity = Integer(secret=True)


class Concert(Table):
    band_1 = ForeignKey(Band)
    band_2 = ForeignKey(Band)
    venue = ForeignKey(Venue)



from __future__ import annotations

from decimal import Decimal
from typing import Callable

import pytest
from litestar import Litestar
from litestar.status_codes import HTTP_200_OK
from litestar.testing import create_test_client
from piccolo.columns import Column, column_types
from piccolo.table import Table
from polyfactory.utils.predicates import is_annotated
from typing_extensions import get_args

from litestar_piccolo import PiccoloDTO, PiccoloPlugin
from tests.app.endpoints import (
    create_concert,
    retrieve_studio,
    retrieve_studio_plugin,
    retrieve_venues,
    retrieve_venues_plugin,
    studio,
    venues,
)
from tests.app.tables import RecordingStudio, Venue

pytestmark = pytest.mark.anyio


def test_dto_serializing_single_piccolo_table(scaffold_piccolo: Callable) -> None:
    with create_test_client(route_handlers=[retrieve_studio]) as client:
        response = client.get("/studio")
        assert response.status_code == HTTP_200_OK
        assert str(RecordingStudio(**response.json()).querystring) == str(studio.querystring)


def test_dto_serializing_multiple_piccolo_tables(scaffold_piccolo: Callable) -> None:
    with create_test_client(route_handlers=[retrieve_venues]) as client:
        response = client.get("/venues")
        assert response.status_code == HTTP_200_OK
        assert [str(Venue(**value).querystring) for value in response.json()] == [str(v.querystring) for v in venues]


def test_plugin_serializing_single_piccolo_table(scaffold_piccolo: Callable) -> None:
    with create_test_client(route_handlers=[retrieve_studio_plugin], plugins=[PiccoloPlugin()]) as client:
        response = client.get("/plugin/studio")
        assert response.status_code == HTTP_200_OK
        assert str(RecordingStudio(**response.json()).querystring) == str(studio.querystring)


def test_plugin_serializing_multiple_piccolo_tables(scaffold_piccolo: Callable) -> None:
    with create_test_client(route_handlers=[retrieve_venues_plugin], plugins=[PiccoloPlugin()]) as client:
        response = client.get("/plugin/venues")
        assert response.status_code == HTTP_200_OK
        assert [str(Venue(**value).querystring) for value in response.json()] == [str(v.querystring) for v in venues]


@pytest.mark.parametrize(
    "piccolo_type, py_type, meta_data_key",
    (
        (column_types.Decimal, Decimal, None),
        (column_types.Numeric, Decimal, None),
        (column_types.Email, str, "max_length"),
        (column_types.Varchar, str, "max_length"),
        (column_types.JSON, str, "format"),
        (column_types.JSONB, str, "format"),
        (column_types.Text, str, "format"),
    ),
)
def test_piccolo_dto_type_conversion(piccolo_type: type[Column], py_type: type, meta_data_key: str | None) -> None:
    class _Table(Table):
        field = piccolo_type(required=True, help_text="my column")

    field_defs = list(PiccoloDTO.generate_field_definitions(_Table))
    assert len(field_defs) == 2
    field_def = field_defs[1]
    assert is_annotated(field_def.raw)
    assert field_def.annotation is py_type
    metadata = get_args(field_def.raw)[1]

    assert metadata.extra.get("description", "")
    if meta_data_key:
        assert metadata.extra.get(meta_data_key, "") or getattr(metadata, meta_data_key, None)


def test_piccolo_dto_openapi_spec_generation(scaffold_piccolo: Callable) -> None:
    app = Litestar(route_handlers=[retrieve_studio, retrieve_venues, create_concert])
    schema = app.openapi_schema

    assert schema.paths
    assert len(schema.paths) == 3
    concert_path = schema.paths["/concert"]
    assert concert_path

    studio_path = schema.paths["/studio"]
    assert studio_path

    venues_path = schema.paths["/venues"]
    assert venues_path

    post_operation = concert_path.post
    assert (
        post_operation.request_body.content["application/json"].schema.ref  # type: ignore
        == "#/components/schemas/CreateConcertConcertRequestBody"
    )

    studio_path_get_operation = studio_path.get
    assert studio_path_get_operation.responses["200"].content["application/json"].schema.ref in {  # type: ignore
        "#/components/schemas/RetrieveStudioRecordingStudioResponseBody",
        "#/components/schemas/tests.app.endpoints.retrieve_studioRecordingStudioResponseBody",
    }

    venues_path_get_operation = venues_path.get
    assert venues_path_get_operation.responses["200"].content["application/json"].schema.items.ref in {  # type: ignore
        "#/components/schemas/RetrieveVenuesVenueResponseBody",
        "#/components/schemas/tests.app.endpoints.retrieve_venuesVenueResponseBody",
    }

    concert_schema = schema.components.schemas["CreateConcertConcertRequestBody"]
    assert concert_schema
    assert concert_schema.to_schema() == {
        "properties": {
            "band_1": {"oneOf": [{"type": "null"}, {"type": "integer"}]},
            "band_2": {"oneOf": [{"type": "null"}, {"type": "integer"}]},
            "venue": {"oneOf": [{"type": "null"}, {"type": "integer"}]},
        },
        "required": [],
        "title": "CreateConcertConcertRequestBody",
        "type": "object",
    }

    record_studio_schema = schema.components.schemas.get(
        "RetrieveStudioRecordingStudioResponseBody",
        schema.components.schemas.get("tests.app.endpoints.retrieve_studioRecordingStudioResponseBody"),
    )
    assert record_studio_schema
    assert record_studio_schema.to_schema() in (
        {
            "properties": {
                "facilities": {"oneOf": [{"type": "null"}, {"type": "string"}]},
                "facilities_b": {"oneOf": [{"type": "null"}, {"type": "string"}]},
                "microphones": {"oneOf": [{"type": "null"}, {"items": {"type": "string"}, "type": "array"}]},
                "id": {"oneOf": [{"type": "null"}, {"type": "integer"}]},
            },
            "required": [],
            "title": "RetrieveStudioRecordingStudioResponseBody",
            "type": "object",
        },
        {
            "properties": {
                "facilities": {"oneOf": [{"type": "null"}, {"type": "string"}]},
                "facilities_b": {"oneOf": [{"type": "null"}, {"type": "string"}]},
                "microphones": {"oneOf": [{"type": "null"}, {"items": {"type": "string"}, "type": "array"}]},
                "id": {"oneOf": [{"type": "null"}, {"type": "integer"}]},
            },
            "required": [],
            "title": "tests.app.endpoints.retrieve_studioRecordingStudioResponseBody",
            "type": "object",
        },
    )

    venue_schema = schema.components.schemas.get(
        "RetrieveVenuesVenueResponseBody",
        schema.components.schemas.get("tests.app.endpoints.retrieve_venuesVenueResponseBody"),
    )
    assert venue_schema
    assert venue_schema.to_schema() in (
        {
            "properties": {
                "capacity": {"oneOf": [{"type": "null"}, {"type": "integer"}]},
                "id": {"oneOf": [{"type": "null"}, {"type": "integer"}]},
                "name": {"oneOf": [{"type": "null"}, {"type": "string"}]},
            },
            "required": [],
            "title": "RetrieveVenuesVenueResponseBody",
            "type": "object",
        },
        {
            "properties": {
                "capacity": {"oneOf": [{"type": "null"}, {"type": "integer"}]},
                "id": {"oneOf": [{"type": "null"}, {"type": "integer"}]},
                "name": {"oneOf": [{"type": "null"}, {"type": "string"}]},
            },
            "required": [],
            "title": "tests.app.endpoints.retrieve_venuesVenueResponseBody",
            "type": "object",
        },
    )



from piccolo.conf.apps import AppRegistry
from piccolo.engine import SQLiteEngine

DB = SQLiteEngine(path=":memory:", check_same_thread=False)

APP_REGISTRY = AppRegistry(
    apps=[
        "tests.app.piccolo_app",
    ],
)
